(* algorithmic core *)
Require Import Ltac2.Ltac2.
Require UFProof.
Require UnionFind.
Require CongruenceClosure.

(* helper to build/run a cc problem *)

Ltac2 Type ('k, 'a, 'c) t := {
    (* *reversed* list of terms *)
    mutable terms : ('k * CongruenceClosure.TermInfo ) list ;
    mutable equations : (int * int * 'a CongruenceClosure.Prf) list ;
    mutable conclusions : (int * int * 'c) list ;
  }.



(* todo: format *)
Ltac2 rec aux_search f lst
:= match lst with
    | [] => None
    | x :: xr => match (f x) with
                | true => Some (List.length xr)
                | false => aux_search f xr
                end
    end.

Ltac2 makeTerm0 (builder: ('k,'a,'c) t) (k: 'k) (data: CongruenceClosure.TermInfo) :=
  let old := builder.(terms) in
  builder.(terms) := (k, data) :: old; 
  List.length old.

Ltac2 makeAtom (builder: ('k,'a,'c) t) (k: 'k) :=
  makeTerm0 builder k CongruenceClosure.Atom.

Ltac2 makeApp (builder: ('k,'a,'c) t) (k: 'k) (f : int) (x: int) :=
  makeTerm0 builder k (CongruenceClosure.App f x).

Ltac2 lookupOr (builder: ('k,'a,'c) t) (query: 'k -> bool) ( fb : unit -> int )
  := match (aux_search (fun (k,_) => query k) (builder.(terms))) with
     | Some i => i
     | None => fb ()
     end.

Ltac2 addEquation builder lhs rhs prf := builder.(equations) := (lhs,rhs,UFProof.axiom lhs rhs prf) :: builder.(equations).

Ltac2 addConclusion builder lhs rhs fn:= builder.(conclusions) := (lhs,rhs,fn) :: builder.(conclusions).


Ltac2 Type ('k, 'a, 'c) Result :=
  [ NoSat('k array, int, int, 'a CongruenceClosure.Prf, 'c) 
  | Sat('k array, ('a CongruenceClosure.CongrAxiom,CongruenceClosure.ClassInfo)UnionFind.t) ].

(* Ltac2 rec find_res_aux (lookup : 'k array) (uf: 'a UnionFind.t) (x: (int * int * 'c) list) :=
  match x with
  | [] => Sat lookup uf
  | tpl :: xr =>
      let (lhs, rhs, concl) := tpl in
      match (UnionFind.congruent uf lhs rhs) with
      | Some prf
      | None => find_res_aux lookup uf x
      end. *)

Ltac2 run (builder: ('k,'a,'c) t) := 
  let uf :=
    CongruenceClosure.findCongruence 
      (Array.of_list (List.rev_map (fun (_,r) => r) (builder.(terms))))
      (builder.(equations)) in
  let lookupArray := (Array.of_list (List.rev_map (fun (l,_) => l) (builder.(terms)))) in
  let rec find_aux lst :=
    match lst with
    | [] => Sat lookupArray uf
    | tpl :: xr =>  
        let (lhs, rhs, concl) := tpl in
        match (UnionFind.congruent uf lhs rhs) with
        | Some prf => NoSat lookupArray lhs rhs prf concl
        | None => find_aux xr
        end
    end in
  find_aux (List.rev (builder.(conclusions))).

Ltac2 new (u: unit) := { terms := [] ; equations := [] ; conclusions := [] }.