Require Import Ltac2.Ltac2.
Require UFProof.
Require CCBuilder.
Require Pick.
Require CongruenceClosure.


Require Import Classes.Morphisms.
Require Import Coq.Relations.Relation_Definitions.
Require Import Setoid.
Require Import Coq.Classes.Morphisms_Prop.

(* draft of working over setoids, has a few problems but overall works well *)
(* needs documenation of both
(1) what it does
and
(2) how it works


(1) will consist mostly of explaining the term structure we produce to simulate the structure of more than one relation
*)


Ltac2 solve_transport 
  (ts: constr) (td: constr)
  (f: constr) (rs: constr) (rd: constr)
    := Control.once (fun () => constr:(fun (x y: $ts) (H: $rs x y) => ltac:(unfold pointwise_relation; setoid_rewrite H; reflexivity) : ($rd ($f x) ($f y)))).

Module helper.
  Polymorphic Instance subr_eq {A: Type} (R: relation A) : Reflexive R -> subrelation eq R.
  Proof.
    intros.
    intros x y [].
    reflexivity.
  Qed.
End helper.


(* relation * term * transport proof *)
Ltac2 Type termType := constr * constr * (constr option).

Ltac2 Type axiomType := [ FromConstr(constr)  ].


(* method adding in coq 8.13, once I can update, use builtin *)
Ltac2 option_get_TEMP (ov : 'a option) :=
  match ov with
  | Some v => v
  | None => Control.throw No_value
  end.

Print transitivity .
Ltac2 rec buildProof (lookup: termType array) (prf: axiomType CongruenceClosure.Prf) 
  := match prf with
      | UFProof.Trans x h y k z 
        => let (rel, xt, _) := Array.get lookup x in
            let (_, yt, _) := Array.get lookup y in
            let (_, zt, _) := Array.get lookup z in
            let l := buildProof lookup h in
            let r := buildProof lookup k in
            constr:(@transitivity _ $rel _ $xt $yt $zt $l $r)
      | UFProof.Sym x y h 
        => let (rel, xt, _) := Array.get lookup x in
            let (_, yt, _) := Array.get lookup y in
            let l := buildProof lookup h in
            constr:(@symmetry _ $rel _ $xt $yt $l)
      | UFProof.Ax y _ a0 => 
        match a0 with
        | CongruenceClosure.Ax a =>
          match a with
          | FromConstr u => u
          end
        | CongruenceClosure.Congr f f' x x' h k
          => let (_, ft, _) := Array.get lookup f in
              let (_, ft', _) := Array.get lookup f' in
              let (rela, xt, _) := Array.get lookup x in
              let (_, xt', _) := Array.get lookup x' in
              let (relb, _, transpInfoMaybe) := Array.get lookup y in
              let transpProof := option_get_TEMP transpInfoMaybe in
              let l := buildProof lookup h in
              let r := buildProof lookup k in
              constr:(@transitivity _ $relb _ ($ft $xt) ($ft $xt') ($ft' $xt') ($transpProof $xt $xt' $r) ($l $xt'))
        end
      | UFProof.Refl x 
        => let (rel, xt, _) := Array.get lookup x in
            constr:(@reflexivity _ $rel _ $xt)
      end.

Ltac2 Type Concs := [ Goal(constr) | Fls(constr) ].

Ltac2 termType_match rel term x 
  := let (rel2, term2, _) := x in 
      match (Constr.equal rel rel2) with
      | false => false
      | true => Constr.equal term term2
      end.
Ltac2 Type exn ::= [ Iterate ].
Ltac2 rec all_values (f: unit -> 't) := 
    match (Control.case f) with
    | Val tpl => let (res, rest) := tpl in res :: all_values (fun _ => rest Iterate)
    | _ => []
    end.

Ltac2 simpleFunctionAppDecomp (x: constr) :=
  lazy_match! x with
  | ?f ?x => lazy_match! (Constr.type f) with
             | ?xt -> ?yt => Some (f,x,xt,yt)
             | _ => None
             end
  | _ => None
  end.

Ltac2 runAll (x: unit -> unit) := Control.plus (fun () => x (); Control.zero Iterate) (fun _ => ()).


Ltac2 c_try (x: unit -> 't) 
  := Control.once (fun () => Control.plus (fun _ => Some (x ())) (fun _ => None)).

Ltac2 addTerm (builder : (termType, axiomType, Concs) CCBuilder.t) (getRelations : constr -> constr) :=
  let rec go rel tm :=
    Control.once (fun () => 
    CCBuilder.lookupOr builder (termType_match rel tm)
    (fun () =>
      match (simpleFunctionAppDecomp tm) with
        | Some tpl0 =>
          let (f,x,xt,yt) := tpl0 in
          match c_try (fun () => let xrel := getRelations xt in (xrel, solve_transport xt yt f xrel rel)) with
          | Some fst =>
            let (xrel,tsp) := fst in
            CCBuilder.makeApp builder (rel,tm,Some tsp) (go constr:(pointwise_relation $xt $rel) f) (go xrel x)
          | _ => CCBuilder.makeAtom builder (rel,tm,None)
          end
        | None => CCBuilder.makeAtom builder (rel,tm,None)
        end
    ))
    in go.



Ltac2 c_last (f: unit -> 'x) :=
  let rec go x0 f := 
    match (Control.case (fun _ => (f Iterate))) with
    | Val tpl =>
        let (fst, rest) := tpl in
        go fst rest
    | Err e => x0
    end
  in 
    match (Control.case f) with
    | Val tpl =>
          let (fst, rest) := tpl in
          go fst rest
    | Err e => Control.zero e
    end.
    

Ltac2 addEquations (builder :  (termType, axiomType, Concs) CCBuilder.t ) (getRelations : constr -> constr) := 
  runAll (fun () =>
    multi_match! goal with
    | [h: ?r ?a ?b |- _] => 
        let xrel := getRelations (Constr.type a) in
        let ht := Control.hyp h in
        let prf := constr:(@is_subrelation _ $r $xrel _ $a $b $ht) in
        CCBuilder.addEquation builder (addTerm builder getRelations xrel a) (addTerm builder getRelations xrel b) (CongruenceClosure.Ax (FromConstr prf))
    end
  ).

Ltac2 get_goal_info (getRelations : constr -> constr) (r: constr) (a: constr) (b: constr)
  := Control.once (fun () => 
        let xrel := getRelations (Constr.type a) in
        (xrel,  constr:(@is_subrelation _ $xrel $r _ $a $b))
     ).

Ltac2 add_goal (builder :  (termType, axiomType, Concs) CCBuilder.t ) (getRelations : constr -> constr) :=
  runAll (fun () =>
  multi_match! goal with
  | [ |- ?r ?a ?b]
             => let (xrel, prf) := get_goal_info getRelations r a b in
                CCBuilder.addConclusion builder (addTerm builder getRelations xrel a) (addTerm builder getRelations xrel b) (Goal prf) 
  | [ h: ~(?r ?a ?b) |- ?r ?a ?b]
             =>  let (xrel, prf) := get_goal_info getRelations r a b in
                 let hp := Control.hyp h in
                 CCBuilder.addConclusion builder 
                  (addTerm builder getRelations xrel a)
                  (addTerm builder getRelations xrel b)
                  (Fls(constr:(fun x => $hp ($prf x))))
  end).



Ltac2 excecute (getRelations : constr -> constr) :=
    let builder := CCBuilder.new () in
    addEquations builder getRelations;
    add_goal builder getRelations;
    match (CCBuilder.run builder) with
    | CCBuilder.Sat _ _ => Control.zero (Tactic_failure (Some (Message.of_string "sat")))
    | CCBuilder.NoSat lookup lhs rhs proof conc 
        => let term := buildProof lookup proof in
          match conc with
          | Goal fn => Control.refine (fun () => constr:($fn $term))
          | Fls(fn) => Control.refine (fun () => ltac1:(exfalso); constr:($fn $term))
          end
    end.

Ltac2 rec try_with_pointwise (getRelations: constr -> constr) (t: constr)
  := multi_match! t with
     | (?x -> ?y) => let next := try_with_pointwise getRelations y in constr:(pointwise_relation $x $next)
     | _ => getRelations x
     end.

Ltac2 tac0 (x: unit -> constr list)
  := Control.enter
      (fun () =>
        let over := x () in 
        excecute (try_with_pointwise (fun t => Pick.from_list over))).

Module Notations.
  Ltac2 Notation "setoid_congruence" rels(thunk(list1(constr, ","))) := tac0 rels.
End Notations.

