Set Universe Polymorphism.


Require Import Ltac2.Ltac2.
Require Import Utf8.
Require Import Coq.Relations.Relation_Definitions.
Require Import Coq.Classes.Morphisms.
Require Import Coq.Classes.Morphisms_Prop.
Require Setoid.
Polymorphic Record EClass {T: Type} (R: relation T) : Type
  := MkEClass {
      repr : T ;
      relation_is_per : PER R ;
      repr_is_proper : Proper R repr
    }.
Arguments EClass {T} R%signature.
Arguments MkEClass {T} {R} repr relation_is_per repr_is_proper.
Arguments repr {T} {R} e.
Arguments relation_is_per {T} {R} e.
Arguments repr_is_proper {T} {R} e.


Definition ClassChildren 
           {A B: Type} {RA: relation A} {RB: relation B}
           (y: EClass RB) (f: EClass (RA ==> RB)) (x: EClass RA) 
           :=  RB (repr f (repr x)) (repr y).

Definition ClassMember {A: Type} {RA: relation A} (c: EClass RA) (v: A) := RA (repr c) v.
Definition ClassEq {A: Type} {RA: relation A} (c1 c2: EClass RA) := RA (repr c1) (repr c2).

Global Instance ClassEq_equiv {A: Type} {RA: relation A}: Equivalence (@ClassEq A RA).
Proof. unfold ClassEq.
  split.
  + intros [ repr ? prop ]. apply prop.
  + intros [ x ? ? ] y. intros. now symmetry.
  + intros [ x ? ? ] y z. apply (@transitivity _ RA _).
Qed.


Global Instance ClassChildrenProper {A B: Type} {RA: relation A} {RB: relation B} : Proper (ClassEq ==> ClassEq ==> ClassEq ==> iff) (@ClassChildren _ _ RA RB).
Proof.
  unfold ClassChildren, ClassEq, Proper. simpl.
  intros [y1 ? ?] [y2 ? ?] yEq
         [f1 ? ?] [f2 ? ?] fEq
         [x1 ? ?] [x2 ? ?] xEq.
  simpl in *.
  apply fEq in xEq.
  ltac1:(setoid_rewrite yEq).
  ltac1:(setoid_rewrite xEq).
  reflexivity.
Qed.








