Set Universe Polymorphism.


Require Import Ltac2.Ltac2.

Module UfProof.
  Ltac2 Type rec 'a t
    := [ Trans ('a t, 'a t)
        | Sym( 'a t )
        (* | CongL( 'a t, int )
        | CongR ( int, 'a t ) *)
        | Ax (int, int, 'a ) 
        | Refl ( int ) ].
  
  Ltac2 sym (x: 'a t)
    := match x with
       | Sym y => y
       | Refl r => Refl r
       | _ => Sym x
       end.
  
  Ltac2 trans (x: 'a t) (y: 'a t)
    := match x with
       | Refl _ => y
       | _ => match y with
              | Refl _ => x
              | _ => Trans x y
              end
        end.
End UfProof.


Module UfNode.
  Ltac2 Type ( 'a ) t
    :=[ Repr ( int ) (* size  *)
      | Node ( 'a UfProof.t, int ) (* proof, parent *)
      ].
End UfNode.

Module Uf.
  Ltac2 Type ( 'a ) t := ('a UfNode.t) array.


  Ltac2 rec find (x : 'a t) (term : int)
    := match (Array.get x term) with
       | UfNode.Repr size => ( term, size, UfProof.Refl term )
       | UfNode.Node proof parent =>
              let ( repTerm, repSize, prf) := find x parent in
              let newProof := UfProof.trans proof prf in
                Array.set x term (UfNode.Node newProof repTerm) ;
                (repTerm, repSize,  newProof)
       end.

  Ltac2 union (x: 'a t) (lhs: int) (rhs: int) (prf: 'a UfProof.t)
    := let (lhsRep, lhsSize, lhsPrf) := find x lhs in 
       let (rhsRep, rhsSize, rhsPrf) := find x rhs in 
       match (Int.equal lhsRep rhsRep) with
       | false =>
            match (Int.ge lhsSize rhsSize) with
            | true => Array.set x lhsRep (UfNode.Repr (Int.add lhsSize rhsSize));
                      Array.set x rhsRep (UfNode.Node 
                                                (UfProof.trans 
                                                (UfProof.trans
                                                (UfProof.sym rhsPrf) (* rhsRep = rhs *)
                                                (UfProof.sym prf)) (* rhs = lhs *)
                                                lhsPrf) (* lhs = lhsRep *)
                                            lhsRep)
            | false => Array.set x lhsRep  (UfNode.Node  
                                        (UfProof.trans 
                                          (UfProof.trans 
                                            (UfProof.sym lhsPrf) (* lhsRep = lhs *)
                                            prf) (* lhs = rhs *)
                                            rhsPrf) (* rhs = rhsRep *)
                                        lhsRep);
                      Array.set x rhsRep (UfNode.Repr (Int.add lhsSize rhsSize))
            end
       | true => ()
       end.

  Ltac2 init (size: int) := Array.make size (UfNode.Repr 1).
End Uf.

(* everything past here needs a re-write *)


Module TermNamer.
  Ltac2 Type t
    := {
      mutable nodes : constr list
    }.

  Ltac2 rec lookupWorker (lst: constr list) (y: constr)
    := match lst with
       | (x :: xr) => match (Constr.equal x y) with
                     | true => (0, true, x :: xr)
                     | false => let (idx, changed, lst2) := lookupWorker xr y in
                                 (Int.add 1 idx, changed, x :: lst2)
                     end
       | [] => (0, false, [y])
    end.
  
  Ltac2 lookupAdd (lst: t) (y: constr)
    := let (idx, changed, newV) := lookupWorker (lst.(nodes)) y in
       lst.(nodes) := newV;
       idx.

  Ltac2 array (x: t) := Array.of_list (x.(nodes)).
  Ltac2 size (x: t) := List.length (x.(nodes)).
  Ltac2 new (x: unit) := { nodes := [] }.
End TermNamer.

Ltac2 capture (u: unit) 
  := 
    let tn := TermNamer.new () in
    let proofs := List.flat_map 
        (fun x => 
            let (ident, _, ty) := x in
            lazy_match! ty with
             | ?x = ?y => [ (Control.hyp ident, TermNamer.lookupAdd tn x, TermNamer.lookupAdd tn y)]
             | _ => []
             end 
            )
        (Control.hyps ()) in
    
    let uf := Uf.init (TermNamer.size tn) in

    List.iter (
      fun x => let (proof, lhs, rhs) := x in
                Uf.union uf lhs rhs (UfProof.Ax lhs rhs proof)
    )  proofs
    ;
      (
        TermNamer.array tn,
        uf
      ).

Goal (forall (T: Type) (x y z: T), x = y -> y = z -> x = z).
Proof. intros.
  Ltac2 Eval (capture ()).



(* 
Module Test.

  Ltac2 testing (u: unit)
    := let uf := Uf.init 5 in
        Uf.union uf 0 1 (UfProof.Ax 0 1 "a1");
        Uf.union uf 2 3 (UfProof.Ax 2 3 "a2");
        Uf.union uf 1 2 (UfProof.Ax 1 2 "a4");
        uf.
  Ltac2 Eval (testing ()).
End Test.


 *)

